// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SURGSIM_TESTING_MLCPIO_TEXTLABELS_H
#define SURGSIM_TESTING_MLCPIO_TEXTLABELS_H

extern const char* const TEXT_LABEL_FLAGS_LIST;
extern const char* const TEXT_LABEL_NUM_DEGREES_OF_FREEDOM;
extern const char* const TEXT_LABEL_NUM_CONSTRAINTS;
extern const char* const TEXT_LABEL_NUM_ATOMIC_CONSTRAINTS;
extern const char* const TEXT_LABEL_CONSTRAINT_TYPES_LIST;
extern const char* const TEXT_LABEL_E_VIOLATIONS_VECTOR;
extern const char* const TEXT_LABEL_HCHt_MLCP_MATRIX;
extern const char* const TEXT_LABEL_MU_FRICTION_VECTOR;
extern const char* const TEXT_LABEL_LAMBDA_VECTOR;
extern const char* const TEXT_LABEL_END_OF_FILE;

#endif // SURGSIM_TESTING_MLCPIO_TEXTLABELS_H
