// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <gtest/gtest.h>

#include "SurgSim/DataStructures/Grid.h"

TEST(Grid, PowerOf3StaticAssertionTest)
{
	static_assert(SurgSim::DataStructures::powerOf3<0>::value == 1, "3^0 != 1");
	static_assert(SurgSim::DataStructures::powerOf3<1>::value == 3, "3^1 != 3");
	static_assert(SurgSim::DataStructures::powerOf3<2>::value == 9, "3^2 != 9");
	static_assert(SurgSim::DataStructures::powerOf3<3>::value == 27, "3^3 != 27");
}
