// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SURGSIM_BLOCKS_DEBUGDUMPBEHAVIOR_H
#define SURGSIM_BLOCKS_DEBUGDUMPBEHAVIOR_H

#include <memory>
#include "SurgSim/Framework/Behavior.h"

namespace SurgSim
{
namespace Input
{

class InputComponent;
}
namespace Graphics
{
class OsgManager;
}

namespace Blocks
{

SURGSIM_STATIC_REGISTRATION(DebugDumpBehavior)

/// Provides keyboard access to debugging functionality
/// F1 - call the graphics manager dumpDebugInfo, this will write the current scenegraph to a file in the working
///      directory.
/// F2 - Display the whole scene and activity status for sceneelements and components
/// F3 - Display tangent generation status for graphics objects
class DebugDumpBehavior : public Framework::Behavior
{

public:

	/// Constructor
	/// \param name
	explicit DebugDumpBehavior(const std::string& name);

	virtual ~DebugDumpBehavior();

	SURGSIM_CLASSNAME(SurgSim::Blocks::DebugDumpBehavior);

	/// Sets the InputComponent, needs to provide keystrokes
	/// \param inputComponent the input component
	void setInputComponent(std::shared_ptr<SurgSim::Framework::Component> inputComponent);

	/// \return the input component that is being used
	std::shared_ptr<SurgSim::Input::InputComponent> getInputComponent() const;

private:

	void update(double dt) override;

	bool doInitialize() override;

	bool doWakeUp() override;

	std::weak_ptr<SurgSim::Graphics::OsgManager> m_manager; ///< Reference to the graphics manager
	std::shared_ptr<SurgSim::Input::InputComponent> m_inputComponent; ///< The input component for reading
	bool m_keyPressedLastUpdate; ///< prevent a repeated keystroke
};

}
}

#endif
