// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "SurgSim/Graphics/MeshPlyReaderDelegate.h"

#include "SurgSim/DataStructures/EmptyData.h"
#include "SurgSim/DataStructures/PlyReader.h"

using SurgSim::DataStructures::PlyReader;

namespace SurgSim
{
namespace Graphics
{

MeshPlyReaderDelegate::MeshPlyReaderDelegate() :
	SurgSim::DataStructures::TriangleMeshPlyReaderDelegate<Mesh>()
{

}


MeshPlyReaderDelegate::MeshPlyReaderDelegate(std::shared_ptr<MeshType> mesh) :
	SurgSim::DataStructures::TriangleMeshPlyReaderDelegate<Mesh>(mesh)

{

}

void MeshPlyReaderDelegate::processVertex(const std::string& elementName)
{
	MeshType::VertexType vertex(SurgSim::Math::Vector3d(m_vertexData.x, m_vertexData.y, m_vertexData.z));

	if (hasTextureCoordinates())
	{
		vertex.data.texture.setValue(SurgSim::Math::Vector2d(m_vertexData.s, m_vertexData.t));
	}

	m_mesh->addVertex(vertex);
}

}
}

