// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "SurgSim/Graphics/OsgTexture.h"

#include <osgDB/ReadFile>

using SurgSim::Graphics::OsgTexture;

OsgTexture::OsgTexture(osg::Texture* texture) :
	m_texture(texture)
{
}

bool OsgTexture::loadImage(const std::string& filePath)
{
	osg::ref_ptr<osg::Image> imageNode = osgDB::readImageFile(filePath);

	if (imageNode.valid())
	{
		m_texture->setImage(0u, imageNode);
		return true;
	}
	else
	{
		return false;
	}
}

void OsgTexture::clearImage()
{
	m_texture->setImage(0, nullptr);
}
