// This file is a part of the OpenSurgSim project.
// Copyright 2013-2015, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SURGSIM_COLLISION_CCDDCDCOLLISION_H
#define SURGSIM_COLLISION_CCDDCDCOLLISION_H

#include "SurgSim/Collision/BoxCapsuleContact.h"
#include "SurgSim/Collision/BoxDoubleSidedPlaneContact.h"
#include "SurgSim/Collision/BoxPlaneContact.h"
#include "SurgSim/Collision/BoxSphereContact.h"
#include "SurgSim/Collision/CapsuleSphereContact.h"
#include "SurgSim/Collision/CompoundShapeContact.h"
#include "SurgSim/Collision/ContactCalculation.h"
#include "SurgSim/Collision/DefaultContactCalculation.h"
#include "SurgSim/Collision/OctreeCapsuleContact.h"
#include "SurgSim/Collision/OctreeDoubleSidedPlaneContact.h"
#include "SurgSim/Collision/OctreePlaneContact.h"
#include "SurgSim/Collision/OctreeSphereContact.h"
#include "SurgSim/Collision/SegmentMeshTriangleMeshContact.h"
#include "SurgSim/Collision/SegmentSelfContact.h"
#include "SurgSim/Collision/SphereDoubleSidedPlaneContact.h"
#include "SurgSim/Collision/SpherePlaneContact.h"
#include "SurgSim/Collision/SphereSphereContact.h"
#include "SurgSim/Collision/TriangleMeshParticlesContact.h"
#include "SurgSim/Collision/TriangleMeshPlaneContact.h"
#include "SurgSim/Collision/TriangleMeshSurfaceMeshContact.h"
#include "SurgSim/Collision/TriangleMeshTriangleMeshContact.h"

#endif // SURGSIM_COLLISION_CCDDCDCOLLISION_H
